// //////////////////////////////////////////////////////////////////////////
//
//     sidechain_pf.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#include "sidechain_pf.h"
#include "getopts.h"   // parse command line args


///
/// main()
///
int main(int argc, char** argv) {
	srand(2);

	///
	/// read input files
	Options shfearOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	shfearOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	shfearOpts.addOption("m",    "map",          "[REQUIRED] Input fs4 map file",      true);
	shfearOpts.addOption("c",    "cryst",        "[REQUIRED] Input crystal file",      true);
	shfearOpts.addOption("s",    "seq",          "[REQUIRED] Input sequence file",      true);
	shfearOpts.addOption("r",    "reso",          "Density map resolution",      true);
	shfearOpts.addOption("B",    "bandwidth",     "Spherical harmonic bandwidth (default 16)",      true);
	shfearOpts.addOption("g",    "rad-align",     "Frag radius for alignment (default 5)",      true);
	shfearOpts.addOption("k",    "rad-mask",      "Frag radius for computing CC (default 8)",      true);
	shfearOpts.addOption("d",    "mask-dist",     "Mask dist for computing CC (default 3)",      true);
	shfearOpts.addOption("N",    "nfrags",        "#fragments per amino acid (default 50)",      true);
	shfearOpts.addOption("p",    "npart",         "Number of particles (default 100)",      true);
	shfearOpts.addOption("b",    "buckets",       "Bucket number (or range) of this particle set (default 1-10)",      true);
	shfearOpts.addOption("y",    "b-min",         "Min belief to continue a trace (default 0.001)",      true);
	shfearOpts.addOption("z",    "M-max",         "Max entropy to start a trace (default 2.8)",      true);
	shfearOpts.addOption("a",    "analogues",     "Structural analogue list for fragment generation",      true);
	shfearOpts.addOption("P",    "partial",       "Partial solution",     true);
	shfearOpts.addOption("G",    "guide",    	 "Input priority function for guiding message passing",      true);

	string workingdir, mapfile, crystfile, seqfile, seleniumfile, partialPDBfile, guidefile;;
	vector< string >  similar_files, bucket_range;
	vector< CAtrace > similar_pdbs;
	CAtrace partialPDB;
	// defaults
	double MASK_DIST=3.0;
	int Ralign=5, Rmax=8, NEXAMPLES=50, NPARTICLES=100, nSimilar=0;
	Bout = B = 16;
	reso = 2.0;
	ENTROPY_CUTOFF = 2.8;
	B_MAX_CUTOFF = 0.001;

	// default to 10 buckets of 100 particles each
	FIRST_BUCKET = 1;
	LAST_BUCKET  = 10;


	shfearOpts.parse(argc, argv);
	while ((switchInt = shfearOpts.cycle()) >= 0) {
		vector<string> pargs;
		Crystal nullCryst;

		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  shfearOpts.getArgs(switchInt); break;
		case 1:   // Option 'map'
			mapfile =  shfearOpts.getArgs(switchInt); break;
		case 2:   // Option 'cryst'
			crystfile =  shfearOpts.getArgs(switchInt); break;
		case 3:   // Option 'seq'
			seqfile =  shfearOpts.getArgs(switchInt); break;
		case 4:   // Option 'reso'
			reso = atof( shfearOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'bandwidth'
			B = Bout = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 6:   // Option 'rad-align'
			Ralign = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 7:   // Option 'rad-mask'
			Rmax = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 8:   // Option 'mask-dist'
			MASK_DIST = atof( shfearOpts.getArgs(switchInt) ); break;
		case 9:   // Option 'nfrags'
			NEXAMPLES = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 10:  // Option 'npart'
			NPARTICLES = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 11:  // Option 'buckets'
			Tokenize( shfearOpts.getArgs(switchInt) , bucket_range , "-" );
			if (bucket_range.size() == 1) {
				FIRST_BUCKET = LAST_BUCKET = atoi(bucket_range[0].c_str());
			} else if (bucket_range.size() == 2) {
				FIRST_BUCKET = atoi(bucket_range[0].c_str());
				LAST_BUCKET = atoi(bucket_range[1].c_str());
			}
			break;
		case 12:  // Option 'b-min'
			B_MAX_CUTOFF = atof( shfearOpts.getArgs(switchInt) ); break;
		case 13:  // Option 'm-max'
			ENTROPY_CUTOFF = atof( shfearOpts.getArgs(switchInt) ); break;
		case 14:  // Option 'analogues'
			Tokenize( shfearOpts.getArgs(switchInt) , similar_files );
			nSimilar = similar_files.size();
			for (int i=0; i<nSimilar; ++i)
				similar_pdbs.push_back( CAtrace( similar_files[i] , nullCryst) );
			break;
		case 15: // Option partial structure
			partialPDBfile = shfearOpts.getArgs(switchInt);
			partialPDB = CAtrace(partialPDBfile,nullCryst,true); break;
		case 16:  //Option 'secondary'
			guidefile = shfearOpts.getArgs(switchInt); break;

		}
	}

	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || mapfile.length() == 0 || crystfile.length() == 0 || seqfile.length() == 0) {
		cerr << "Error: Must specify a working directory, fs4 map file, crystal file, and sequence file.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}

	// read input files
	map_3d<double> origMap;
	Crystal crystData;
	readDensityMap(origMap, mapfile, crystData, crystfile);
	Sequence seqData(seqfile);
	string filestem = seqfile;
		filestem = filestem.substr(filestem.find_last_of('/')+1);
		filestem = filestem.substr(0,filestem.find_last_of('.'));
	string outdir = workingdir + "/" + itoc(NPARTICLES) + "particle_models/";
	string marginalMapsPath = workingdir + "/posteriors/";

	mkdir(outdir.c_str(),    0755);   // ensure dirs exists

	int nResSum = seqData.nres(), NFRAGS;

	if (Ralign > Rmax) {
		cerr << "WARNING: Ralign > Rmax\n";
		Ralign=Rmax;
	}
	Rstep = 1.0;

	chem_db_initialize();

	//////////////////////
	///
	///  pretty-print
	cerr << "Running with inputs: " << endl;
	cerr << "    working directory = \'" << workingdir << "\'" << endl;
	cerr << "    fs4-format map    = <" << mapfile << ">" << endl;
	cerr << "    crystal dims      = " << crystData.to_string() << endl;
	cerr << "            spacegp   = " << crystData.symmGpId << endl;
	cerr << "    protein sequence  = " << seqData.to_string() << endl;
	cerr << "    resolution        = " << reso << endl;

	cerr << endl;
	cerr << "Search parameters " << endl;
	cerr << "    B                 = " << B << endl;
	cerr << "    radius-align      = " << Ralign << endl;
	cerr << "    radius-mask       = " << Rmax << endl;
	cerr << "    mask-dist         = " << MASK_DIST << endl;
	cerr << "    nfrags            = " << NEXAMPLES << endl;
	cerr << "    nparticles        = " << NPARTICLES << endl;
	cerr << "    min_belief        = " << B_MAX_CUTOFF << endl;
	cerr << "    max_entropy       = " << ENTROPY_CUTOFF << endl;


	cerr << endl;
	if (nSimilar > 0) {
		cerr << "Using similar maps: " << endl;
		for (int i=0; i<nSimilar; ++i)
			cerr << "    " << similar_files[i] << endl;
	}


	/////////////////////////
	///
	/// particles
	vector< Particle >   part(NPARTICLES , Particle(nResSum, crystData) );
	vector< Posterior >  b(nResSum);
	//Posterior b_i;

	//vector< double >     entropy_byPosition (nResSum, INFINITY);
	vector< double >     entropy_byPosterior(nResSum, INFINITY);
	vector< double >     prob_byPosition (nResSum, 0.0);
	vector< double >     prob_byPosterior(nResSum, 0.0);

	vector< int >        b_cover_sum(nResSum,0);


	///
	/////////////////////////
	vector<int> guidefxn(nResSum,0);
	if(guidefile.length() > 0){
		Sequence guideSeq;
		guideSeq.load(guidefile);
		if(guideSeq.nres() != nResSum){
			cerr << "Error: Guiding function do not correspond to sequence length" << endl;
			exit(1);
		}
		for(int i_res = 0 ; i_res < nResSum; i_res++){
			char val = guideSeq.serialGet(i_res);
			guidefxn[i_res] = atoi(&val);
		}
	}

	/////////////////////////
	///
	/// load posterior maps from disk
	/// compute information content to see where to begin

	map_3d<double> map , b_full;
	int resId, startingRes = -1;
	char resType, chnId;
	vec3<int> dims;
	for (int i_res = 0; i_res < nResSum; i_res++) {
		string mer = seqData.serialGet5mer(i_res);
		if (mer[2] == '_'  || part[0].b_cover[ i_res ]) continue;  // gap or partial structure
		seqData.serialGet(i_res, resId, resType, chnId);
		string marginalFileName = marginalMapsPath+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal.mat";

		if (!exists(marginalFileName))
			marginalFileName = marginalMapsPath+"prot_chn"+chnId+"_res"+itoc(resId)+"_marginal.mat";

		b[i_res] = Posterior(marginalFileName, mer, crystData, dims);
		entropy_byPosterior [i_res] = b[i_res].M.entropy();
	}

	for (int i=0; i<nResSum; ++i)
		prob_byPosterior[i] = exp( - entropy_byPosterior[i] ) ;
	for(int res = 0; res < partialPDB.nres(); res++){
		partialPDB.serialGet(res, resId, resType, chnId);;
		if(resType == '_') continue;
		prob_byPosterior[resId-1] = 0.0;
	}
	ENTROPY_CUTOFF = exp( - ENTROPY_CUTOFF );

	map.alloc(dims[0],dims[1],dims[2]);
	resize_fft(origMap, map); // resample map
	//
	///////////////////////

	///////////////////////
	///
	/// spherical harmonic stuff
	map_3d< cubicSplineCoeff3d > sphPtrsMap(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsMap, dims, crystData);
	vector< map_3d<double> > coeffsR, coeffsI, maskCoeffsR, maskCoeffsI;

	string fragmentDir = workingdir + "/fragment_cache/";
	mkdir(fragmentDir.c_str(),    0755);   // ensure dir exists
	vector< Nmer > fragments;

	SOFT_initialize();  // initializes temp space for spharms
	///
	/////////////////////////


	///////////////////////
	///
	/// Prob by position
	for (int i_res = 0; i_res < nResSum; i_res++) {
		if (prob_byPosterior[i_res] == 0.0) continue;

		prob_byPosition[i_res] = 0.0;
		vector<int> altReses = seqData.serialGetCopies(i_res);

		for (int i_cpy = 0; i_cpy < altReses.size(); ++i_cpy) {
			prob_byPosition[i_res] += prob_byPosterior[ altReses[i_cpy] ];
		}
		prob_byPosition[i_res] /= altReses.size();
	}

	vector<double> orig_prob_byPosition    = prob_byPosition;
	vector<int> orig_guidefxn = guidefxn;
	///
	/////////////////////////


	for (int BUCKET_ID = FIRST_BUCKET; BUCKET_ID <= LAST_BUCKET; ++BUCKET_ID) {
		cerr << "[BUCKET " << BUCKET_ID << "]\n\n";

		srand(BUCKET_ID);  // Consistency w/i runs from the same bucket

		//////////////////////////
		///
		/// "reset" particles
		part.clear();
		part.resize(NPARTICLES , Particle(nResSum, crystData) );
		prob_byPosition    = orig_prob_byPosition;
		guidefxn = orig_guidefxn;
		b_cover_sum.clear();
		b_cover_sum.resize(nResSum,0);
		///
		//////////////////////////
		for(int res = 0; res < partialPDB.nres(); res++){
			char resType, chnId;
			int resId;
			partialPDB.serialGet(res, resId, resType, chnId);
			if(resType == '_') continue;
			Nmer frag_taken = Nmer(partialPDBfile,chnId,resId,0.0,5);
			PDBResidue Calpha = partialPDB.serialGet(res);
			frag_taken.translate(Calpha.cartX);

			Residue thisRes;
			frag_taken.getCenterRes(thisRes);

			thisRes.CAm1 = frag_taken.getAtom(1," CA ");
			thisRes.N    = frag_taken.getAtom(2," N  ");
			thisRes.CA   = frag_taken.getAtom(2," CA ");
			thisRes.C    = frag_taken.getAtom(2," C  ");
			thisRes.O    = frag_taken.getAtom(2," O  ");
			thisRes.CAp1 = frag_taken.getAtom(3," CA ");
			thisRes.restype = AA2ID[ resType ];
			for(int i = 0; i < NPARTICLES; i++){
				part[i].trace[resId-1] = thisRes;
				part[i].b_cover[ resId-1 ] = true;
				b_cover_sum[ resId-1 ]++;
				part[i].mask.updateMask( resId-1 , thisRes.atoms , crystData);
				part[i].trace[resId-1].b = 1;
			}
		}

		///////////////////////////
		///
		/// choose start
		double maxProb = 0.0, probSum = 0.0;
		int maxGuide = 0, guideSum = 0;
		for (int i_res = 1; i_res < nResSum-1; i_res++) {
			if (    prob_byPosition[i_res] > ENTROPY_CUTOFF
				 && prob_byPosition[i_res-1] != 0.0 && prob_byPosition[i_res+1] != 0.0 ) {
				probSum += prob_byPosition[i_res];
				guideSum += guidefxn[i_res];

				if( guidefxn[i_res] > maxGuide ||
					(guidefxn[i_res] == maxGuide && prob_byPosition[i_res] > prob_byPosition[startingRes])) {
					maxGuide = guidefxn[i_res];
					maxProb = prob_byPosition[i_res];
					startingRes = i_res;
				}
			}
		}

		if (BUCKET_ID != 1) {
			double resChooser = (double)rand() / RAND_MAX;
			resChooser *= (guidefile.length() > 0) ? guideSum : probSum;
			for (int i_res = 1; i_res < nResSum-1; i_res++) {
				if (    prob_byPosition[i_res] > ENTROPY_CUTOFF
				     && prob_byPosition[i_res-1] != 0.0 && prob_byPosition[i_res+1] != 0.0 ) {
					resChooser -= (guidefile.length() > 0) ? guidefxn[i_res] : prob_byPosition[i_res];
					if (resChooser <= 0.0) {
						startingRes = i_res;
						break;
					}
				}
			}
		}
		///
		//////////////////////////

		//////////////////////////
		///
		/// main loop
		///
		//while (minEntropy < ENTROPY_CUTOFF) {  // choose some other entropy cutoff????
		while (maxProb >= ENTROPY_CUTOFF) {  // choose some other entropy cutoff????
			string mer = seqData.serialGet5mer(startingRes);

			//lookup
			fragment_lookup( fragmentDir, mer, Rmax, Ralign, NEXAMPLES, MASK_DIST,
			                 coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
			                 fragments, similar_files, similar_pdbs);
			NFRAGS = fragments.size();

			// place starting res
			vector<int> altReses = seqData.serialGetCopies(startingRes);

			//////////////////////
			///
			/// START the trace
			///
			cerr << "Starting new chain trace at residue " << startingRes << "." << endl;
			int nnonzero = 0;
			vector< double > oldWeights(NPARTICLES, 0.0);

			for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
				part[i_prt].trace[startingRes].restype = AA2ID[ mer[2] ];  // set restype

				double minEntropy = INFINITY;
				int startingPosterior;
				for (int i_cpy=0; i_cpy<altReses.size(); ++i_cpy) {
					if (    part[i_prt].b_cover[ altReses[i_cpy] ] == false
						&& entropy_byPosterior[ altReses[i_cpy] ] < minEntropy ) {
						minEntropy = entropy_byPosterior[ altReses[i_cpy] ];
						startingPosterior = altReses[i_cpy];
					}
				}
				compact_map_3d<float> *bPtr = &(b[ startingPosterior ].M);
				part[i_prt].b_cover[ startingPosterior ] = true;
				b_cover_sum[ startingPosterior ]++;

				oldWeights[i_prt] = part[i_prt].weight;
				startTrace( part[i_prt].trace[startingRes], bPtr, map,
							fragments, coeffsR, coeffsI,
							Ralign, crystData, part[i_prt].weight, part[i_prt].mask, startingRes );

				if (part[i_prt].weight > 0.0)
					nnonzero++;
			}
			cerr << endl;
			//entropy_byPosition[startingRes] = INFINITY;
			prob_byPosition[startingRes] = 0.0;
			guidefxn[startingRes] = 0;

			// extend in both directions as far as possible
			int leftFrontier  = startingRes-1, rightFrontier = startingRes+1, currFrontier, prevFrontier, prevprevFrontier;
			bool continueLeftExtension=true, continueRightExtension=true;
			int extensionDir;

			double weightSum, weightMax, b_sum_considered, b_considered;

			string merL=seqData.serialGet5mer(leftFrontier+1), merR=seqData.serialGet5mer(rightFrontier-1);
			//continueLeftExtension  &= (merL[1] != '_') && (entropy_byPosition[leftFrontier]  != INFINITY);
			//continueRightExtension &= (merR[3] != '_') && (entropy_byPosition[rightFrontier] != INFINITY);
			continueLeftExtension  &= (merL[1] != '_') && (prob_byPosition[leftFrontier]  != 0.0);
			continueRightExtension &= (merR[3] != '_') && (prob_byPosition[rightFrontier] != 0.0);

			if (nnonzero == 0) {
				cerr << "Unable to place residue" << startingRes << endl ;
				cerr << "  (no particle with nonzero weight)\n";
				continueRightExtension = continueLeftExtension = false;

				// restore old weights
				for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt)
					part[i_prt].weight = oldWeights[i_prt];
			}


			////////////////////////
			///
			/// EXTEND the trace
			///
			while (continueLeftExtension || continueRightExtension) {
				int currAAid;
				vec3<double> CAm1, CA0, CAp1;
				Residue tempRes;

				// pick a direction based on entropy
				if (continueLeftExtension && continueRightExtension) {
					//extensionDir = entropy_byPosition[leftFrontier] < entropy_byPosition[rightFrontier] ? -1 : +1;
					extensionDir = ((guidefxn[leftFrontier] > guidefxn[rightFrontier]) ||
									(guidefxn[leftFrontier] == guidefxn[rightFrontier]) && (prob_byPosition[leftFrontier] > prob_byPosition[rightFrontier]))
									? -1 : +1;
					if (BUCKET_ID != 1) {
						double dirPicker = ((double)rand() / RAND_MAX);
						dirPicker *= (guidefile.length() > 0) ? (guidefxn[rightFrontier]+guidefxn[leftFrontier]) :(prob_byPosition[leftFrontier] + prob_byPosition[rightFrontier]);
						if(guidefile.length()>0)
							extensionDir = guidefxn[leftFrontier] > dirPicker ? -1 : +1;
						else
							extensionDir = prob_byPosition[leftFrontier] > dirPicker ? -1 : +1;
					}
				} else {
					extensionDir = continueLeftExtension ? -1 : +1;
				}

				// direction-specific variables
				if (extensionDir>0) {
					mer = merR;
					currFrontier = rightFrontier;
					prevFrontier = rightFrontier-1;
					prevprevFrontier = rightFrontier-2;
					currAAid = AA2ID[ mer[3] ];
				} else {
					mer = merL;
					currFrontier = leftFrontier;
					prevFrontier = leftFrontier+1;
					prevprevFrontier = leftFrontier+2;
					currAAid = AA2ID[ mer[1] ];
				}

				cerr << "Placing residue " << currFrontier << " for all particles.\n";
				prob_byPosition[currFrontier] = 0.0;
				guidefxn[currFrontier] = 0;
				///////////////////////
				///
				/// (A) select p(bb_k | bb_{k-1})
				///
				weightMax = 0.0;
				vector<int> altReses = seqData.serialGetCopies(currFrontier);
				vector< compact_map_3d<float>* > bs;
				vector< int > is;
				int nalt = altReses.size();

				for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
					part[i_prt].trace[currFrontier].restype = currAAid;  // set restype

					bs.clear(); is.clear();
					for (int i_cpy=0; i_cpy<nalt; ++i_cpy) {
						if (  part[i_prt].b_cover[ altReses[i_cpy] ] == false  ) {
							bs.push_back( &(b[ altReses[i_cpy] ].M) );
							is.push_back( altReses[i_cpy] );
						}
					}

					int chosenPosteriorIdx =
						forward_sample_CA(extensionDir, bs, crystData,
										part[i_prt].trace[prevFrontier],
										part[i_prt].trace[currFrontier],
										b_sum_considered,
										b_considered);  // sets thisRes.CA

					// update cover
					part[i_prt].trace[currFrontier].b = b_considered;

					int chosenPosterior = is[chosenPosteriorIdx];
					part[i_prt].b_cover[ chosenPosterior ] = true;
					b_cover_sum[ chosenPosterior ]++;

					part[i_prt].weight *= b_sum_considered ;

					weightMax = max(weightMax, b_considered);
				}

				// if wtMax is too low, abort trace
				if (weightMax < B_MAX_CUTOFF) {
					// no match for last res
					if (extensionDir<0) {
						cerr << "Aborting LEFT extension at " << currFrontier << endl
						     << "  (no particle with probability > " << B_MAX_CUTOFF << ")" << endl;
						continueLeftExtension = false;
					} else {
						cerr << "Aborting RIGHT extension at " << currFrontier << endl
						     << "  (no particle with probability > " << B_MAX_CUTOFF << ")" << endl;
						continueRightExtension = false;
					}
					continue;
				}


				// if the prev res was the seed, don't to add a sidechain (already added)
				if (part[0].trace[prevFrontier].atoms.size() == 0) {
					fragment_lookup( fragmentDir, mer, Rmax, Ralign, NEXAMPLES, MASK_DIST,
									coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
									fragments, similar_files, similar_pdbs);
					NFRAGS = fragments.size();

					///////////////////////
					///
					/// (B) compute CC( s_{k-1,j} , map(bb_{k-1}) )
					///        == corrcoef of map around b_{k-1} with each possible sidechain (k-1)'s density
					///
					vector< vector< double > > traceCCs(NPARTICLES , vector< double >( NFRAGS , -INFINITY ) );
					vector< vector< double > > traceProbs(NPARTICLES , vector< double >( NFRAGS ) );
					vector< vector< mat33<double> > > traceRots(NPARTICLES , vector< mat33<double> >( NFRAGS ) );

					for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
						if (extensionDir>0) {
							CAm1 = part[i_prt].trace[prevFrontier].CAm1;
							CA0  = part[i_prt].trace[prevFrontier].CA;
							CAp1 = part[i_prt].trace[currFrontier].CA;
						} else {
							CAm1 = part[i_prt].trace[currFrontier].CA;
							CA0  = part[i_prt].trace[prevFrontier].CA;
							CAp1 = part[i_prt].trace[prevFrontier].CAp1;
						}
						matchSidechains( CAm1,CA0,CAp1, map,
										fragments, coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
										crystData, traceCCs[i_prt], traceRots[i_prt]);
					}

					///////////////////////
					///
					/// (C) compute prob from CC's
					///        - compute prob(CC) in 'traceProbs'
					///        - thisTraceWts holds prob sum (will be used to update traceWts)
					///
					//cerr << "c\n";
					double CCmean  = mean(traceCCs);
					double CCstdev = sqrt(variance(traceCCs));
					vector< double > thisTraceWts(NPARTICLES, 0.0);
					nnonzero = 0;

					for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
						Residue &prevRes = part[i_prt].trace[prevFrontier];

						for (int i_frag=0; i_frag<NFRAGS; ++i_frag) {
							double pNull = normCDF( (CCmean - traceCCs[i_prt][i_frag]) / CCstdev);
							traceProbs[i_prt][i_frag] += (1.0 - pNull) / pNull;

							// collision ==> p=0
							fragments[i_frag].getCenterRes(tempRes);
							tempRes.rotate( traceRots[i_prt][i_frag] );
							tempRes.translate( prevRes.CA );

							// check for collision
							if ( part[i_prt].mask.checkMask( prevFrontier, tempRes.atoms , crystData) )
								traceProbs[i_prt][i_frag] = 0.0;

							thisTraceWts[i_prt] += traceProbs[i_prt][i_frag];
						}

						oldWeights[i_prt] = part[i_prt].weight;
						part[i_prt].weight *= thisTraceWts[i_prt];   // UPDATE sidechain wt

						if (part[i_prt].weight > 0.0) {
							nnonzero++;

							// select sidechain conf with prob \propto traceCCs[i_prt][i_frag]
							double fragPicker = thisTraceWts[i_prt] * ((double)rand())/RAND_MAX;
							int frag_taken;
							double prob_taken, cc_taken;
							mat33<double> rot_taken;
							for (int i_frag=0; i_frag<NFRAGS; ++i_frag) {
								fragPicker -= traceProbs[i_prt][i_frag];
								if (fragPicker <= 0) {
									frag_taken = i_frag;
									prob_taken = traceProbs[i_prt][i_frag] / thisTraceWts[i_prt];
									cc_taken   = traceCCs[i_prt][i_frag];
									rot_taken  = traceRots[i_prt][i_frag];
									break;
								}
							}

							// We've picked a sidechain
							// Now extract atoms, rotate, translate, update mask
							fragments[frag_taken].getCenterRes(prevRes);

							prevRes.rotate(rot_taken);
							prevRes.translate(prevRes.CA);
							prevRes.N   = fragments[frag_taken].getAtom(2," N  ").apply(rot_taken) + prevRes.CA;
							prevRes.C   = fragments[frag_taken].getAtom(2," C  ").apply(rot_taken) + prevRes.CA;
							prevRes.O   = fragments[frag_taken].getAtom(2," O  ").apply(rot_taken) + prevRes.CA;

							// check distances
							part[i_prt].mask.updateMask( prevFrontier , prevRes.atoms , crystData );
						}
					}

					cerr << "Currently " << nnonzero << " particles with nonzero weight.\n\n";

					if (nnonzero == 0) {
						// no match for last res
						if (extensionDir<0) {
							cerr << "Aborting LEFT extension at " << currFrontier << endl ;
							cerr << "  (no particle with nonzero weight)\n";
							continueLeftExtension = false;
						} else {
							cerr << "Aborting RIGHT extension at " << currFrontier << endl ;
							cerr << "  (no particle with nonzero weight)\n";
							continueRightExtension = false;
						}
						// restore old weights
						for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt)
							part[i_prt].weight = oldWeights[i_prt];

						continue;
					}

					// resample! -- select NPARTICLES _with_ replacement
					resample_traces(part, b_cover_sum);
				}



				/////////////////////////
				///
				/// (D) next residue
				///
				//cerr << "d\n";
				bool isTerminus;
				if (extensionDir>0) {
					merR = seqData.serialGet5mer(rightFrontier);
					rightFrontier++;
					isTerminus = (merR[3] == '_') || (prob_byPosition[rightFrontier] == 0.0);
					continueRightExtension &= (!isTerminus);

					if (isTerminus) {
						mer = merR;
						currFrontier = rightFrontier-1;
						prevFrontier = rightFrontier-2;
						prevprevFrontier = rightFrontier-3;
					}
				} else {
					merL = seqData.serialGet5mer(leftFrontier);
					leftFrontier--;
					isTerminus = (merL[1] == '_') || (prob_byPosition[leftFrontier]  == 0.0);
					continueLeftExtension  &= (!isTerminus);

					if (isTerminus) {
						mer = merL;
						currFrontier = leftFrontier+1;
						prevFrontier = leftFrontier+2;
						prevprevFrontier = leftFrontier+3;
					}
				}

				if (isTerminus) {
					cerr << "Terminating (normally) extension at " << currFrontier << endl;
					fragment_lookup( fragmentDir, mer, Rmax, Ralign, NEXAMPLES, MASK_DIST,
									coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
									fragments, similar_files, similar_pdbs);
					NFRAGS = fragments.size();
					nnonzero = 0;

					for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
						oldWeights[i_prt] = part[i_prt].weight;
						matchTerminus(extensionDir,
									part[i_prt].trace[currFrontier],
									part[i_prt].trace[prevFrontier].CA,
									(extensionDir>0) ? part[i_prt].trace[prevFrontier].CAm1 : part[i_prt].trace[prevFrontier].CAp1,
									map, fragments, coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
									crystData, part[i_prt].weight, part[i_prt].mask, currFrontier );

						if (part[i_prt].weight > 0.0)
							nnonzero++;
					}

					if (nnonzero == 0) {
						// restore old weights
						for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt)
							part[i_prt].weight = oldWeights[i_prt];
					}
					cerr << "Currently " << nnonzero << " particles with nonzero weight.\n\n";
				}
			}
			//break;

			///////////////////
			///
			/// recompute probs
			for (int i_res = 0; i_res < nResSum; i_res++) {
				if (prob_byPosition[i_res] == 0.0) continue;

				prob_byPosition[i_res] = 0.0;
				vector<int> altReses = seqData.serialGetCopies(i_res);

				int b_cover_sumSum = 0;
				for (int i_cpy = 0; i_cpy < altReses.size(); ++i_cpy) {
					prob_byPosition[i_res] += (NPARTICLES - b_cover_sum[ altReses[i_cpy] ]) * prob_byPosterior[ altReses[i_cpy] ];
					b_cover_sumSum += b_cover_sum[ altReses[i_cpy] ];
				}
				prob_byPosition[i_res] /= (NPARTICLES*altReses.size() - b_cover_sumSum);
			}

			///////////////////
			///
			/// find a new seed
			maxProb = probSum = 0.0;
			maxGuide =  guideSum = 0;
			for (int i_res = 1; i_res < nResSum-1; i_res++) {
				if (     prob_byPosition[i_res] > ENTROPY_CUTOFF
				      && prob_byPosition[i_res-1] != 0.0 && prob_byPosition[i_res+1] != 0.0 ) {
					probSum += prob_byPosition[i_res];
					guideSum += guidefxn[i_res];

					if( guidefxn[i_res] > maxGuide ||
						(guidefxn[i_res] == maxGuide && prob_byPosition[i_res] > prob_byPosition[startingRes])) {
						maxGuide = guidefxn[i_res];
						maxProb = prob_byPosition[i_res];
						startingRes = i_res;
					}
				}
			}

			if (BUCKET_ID != 1) {
				double resChooser = (double)rand() / RAND_MAX;
				resChooser *= (guidefile.length() > 0) ? guideSum : probSum;
				for (int i_res = 1; i_res < nResSum-1; i_res++) {
					if (    prob_byPosition[i_res] > ENTROPY_CUTOFF
						 && prob_byPosition[i_res-1] != 0.0 && prob_byPosition[i_res+1] != 0.0 ) {
						resChooser -= (guidefile.length() > 0) ? guidefxn[i_res] : prob_byPosition[i_res];
						if (resChooser <= 0.0) {
							startingRes = i_res;
							break;
						}
					}
				}
			}
		}


		//////////
		///
		/// find highest weight particle
		int bestPart=-1;
		double bestPartWt=0.0;
		for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
			if (part[i_prt].weight > bestPartWt) {
				bestPartWt = part[i_prt].weight;
				bestPart = i_prt;
			}
		}

		string filename = outdir+filestem+"_raw0_pf."+itoc(BUCKET_ID)+".pdb";
		writeParticleToFile(filename, part[bestPart], seqData, crystData);

		filename = outdir+filestem+"_allatom_pf."+itoc(BUCKET_ID)+".pdb";
		findSingleAsymmUnit(part[bestPart], seqData, crystData);
		writeParticleToFile(filename, part[bestPart], seqData, crystData);
	}
}
